"""provides a function that parses the command line parameters"""
import argparse


def get_parameters():
    """parses the parameters from command line"""
    parser = argparse.ArgumentParser(
        description="Please provide at least one of the following: "
        "1) a Azure Blob Storage name OR 2) a connection string. "
    )
    parser.add_argument(
        "-c",
        "--connectionstring",
        help="Azure Blob Storage connection string.",
    )
    parser.add_argument(
        "-s",
        "--storage",
        help="Azure Blob Storage resource name. Creates new one if not existing.",
    )
    parser.add_argument(
        "-r",
        "--resourcegroup",
        help="The Azure Blob Storage is in this resource group. Default: blogs-rg",
        default="blogs-rg",
    )
    parser.add_argument(
        "-f",
        "--folder",
        help="Folder with static website data. Will be pushed to the storage.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Verbose, use this flag for debugging.",
        action="store_true",
    )
    return parser.parse_args()



"""provides a container for credentials"""
from getpass import getpass
import re


def get_info_from_cli(name, hidden=False):
    """gets an information from command line"""
    if hidden:
        info = getpass(f"Enter {name}: ")
    else:
        info = input(f"Enter {name}: ")
    return info


def is_connection_string(connectionString):
    """checks if a string is a valid azure blob connection string"""
    if "AccountName" in connectionString and "AccountKey" in connectionString:
        return True
    return False


class Credentials:
    """holds the required credentials for the tool, gets data from CLI
    if they are not available"""

    def __init__(self, args):
        try:
            self.connectionstring = args.connectionstring
        except AttributeError:
            self.connectionstring = None
        try:
            self.folder = args.folder
        except AttributeError:
            self.folder = None

        # get Connection String
        if self.connectionstring is None:
            self.connectionstring = get_info_from_cli(name="connection string")
            if not is_connection_string(self.connectionstring):
                raise ValueError("connection string is invalid!")
        self.storageName = self.get_value_from_connectionstring("AccountName")
        self.storageKey = self.get_value_from_connectionstring("AccountKey")

        # get local folder
        if self.folder is None:
            self.folder = get_info_from_cli(name="local folder")

    def get_value_from_connectionstring(self, value):
        """returns one of these values from the connection string:
        DefaultEndpointsProtocol, AccountName, AccountKey, EndpointSuffix"""
        return re.findall(f"{value}=([^;]+)", self.connectionstring)[0]



"""template for azure"""
TEMPLATE = {
    "$schema": "https://schema.management.azure.com/schemas/2019-04-01/deploymentTemplate.json#",
    "contentVersion": "1.0.0.0",
    "parameters": {
        "storageAccounts_blogdatastore_name": {
            "defaultValue": "blogdatastore",
            "type": "String",
        }
    },
    "variables": {},
    "resources": [
        {
            "type": "Microsoft.Storage/storageAccounts",
            "apiVersion": "2022-05-01",
            "name": "[parameters('storageAccounts_blogdatastore_name')]",
            "location": "westeurope",
            "sku": {"name": "Standard_LRS", "tier": "Standard"},
            "kind": "StorageV2",
            "properties": {
                "dnsEndpointType": "Standard",
                "defaultToOAuthAuthentication": False,
                "publicNetworkAccess": "Enabled",
                "allowCrossTenantReplication": True,
                "routingPreference": {
                    "routingChoice": "InternetRouting",
                    "publishMicrosoftEndpoints": False,
                    "publishInternetEndpoints": False,
                },
                "minimumTlsVersion": "TLS1_0",
                "allowBlobPublicAccess": True,
                "allowSharedKeyAccess": True,
                "networkAcls": {
                    "bypass": "AzureServices",
                    "virtualNetworkRules": [],
                    "ipRules": [],
                    "defaultAction": "Allow",
                },
                "supportsHttpsTrafficOnly": True,
                "encryption": {
                    "requireInfrastructureEncryption": False,
                    "services": {
                        "file": {"keyType": "Account", "enabled": True},
                        "blob": {"keyType": "Account", "enabled": True},
                    },
                    "keySource": "Microsoft.Storage",
                },
                "accessTier": "Hot",
            },
        }
    ],
}



"""executable part of the module"""
import logging
from absu.modules.parameterparser import get_parameters
from absu.modules.credentials import Credentials
from absu.modules.azureblobconnector import AzureBlobConnector
from absu.modules.azureconnector import AzureConnector


if __name__ == "__main__":
    logging.info("Getting parameters from parameters...")
    args = get_parameters()
    logging.info("Ok.")
    if args.verbose:
        logging.root.setLevel(logging.NOTSET)
    if not args.connectionstring:
        logging.info("No connection string given, connecting to Azure...")
        ac = AzureConnector()
        logging.info("Ok.")
        if not args.storage:
            args.storage = input("Azure Blob Storage account name: ")
        logging.info("Creating/updating resource group...")
        ac.create_resource_group(args.resourcegroup)
        logging.info("Ok.")
        logging.info("Creating/updating storage...")
        while not ac.create_storage_account(args.resourcegroup, args.storage):
            ac.storage = input("Could not create storage, try another name: ")
        logging.info("Ok.")
        logging.info("Downloading connection string...")
        args.connectionstring = ac.get_connection_string(
            args.storage, args.resourcegroup
        )
        logging.info("Ok.")
    logging.info("Checking foldername + connection string credentials...")
    cred = Credentials(args)
    logging.info("Ok.")
    logging.info("Connecting to storage...")
    abc = AzureBlobConnector(cred)
    logging.info("Ok.")
    logging.info("Creating container on storage...")
    abc.create_container()  # if doesnt exist
    logging.info("Ok.")
    logging.info("Deleting all files in container...")
    abc.clean_container()  # if files exist
    logging.info("Ok.")
    logging.info("Uploading folder to storage...")
    abc.upload_folder()  # files in folder



"""provides connectors for Azure Blob"""
import os
import logging
from azure.storage.blob import BlobServiceClient, ContentSettings
import magic


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer."""
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError(f"invalid default answer: '{default}'")
    while True:
        print(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        if choice in valid:
            return valid[choice]
        print("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")


class AzureBlobConnector:
    """Handles Azure Blob Storage communication"""

    def __init__(self, creds):
        self.container = "$web"
        self.creds = creds
        self.bsc = BlobServiceClient.from_connection_string(
            self.creds.connectionstring
        )
        self.containerClient = self.bsc.get_container_client(
            container=self.container
        )

    def create_container(self):
        """makes sure the container exists. True if didn't exist else False"""
        returnvalue = False
        containers = []
        for container in self.bsc.list_containers():
            containers.append(container["name"])
        if self.container not in containers:
            self.bsc.create_container(self.container, public_access="BLOB")
            returnvalue = True
        return returnvalue

    def clean_container(self):
        """deletes files in container"""
        if query_yes_no(f"Delete all files in the container {self.container}"):
            for i in self.containerClient.list_blobs():
                self.containerClient.delete_blob(blob=i["name"])
        else:
            raise SystemExit("User exited")

    def upload_folder(self):
        """uploads the folder to the container"""
        for root, _, files in os.walk(self.creds.folder):
            for name in files:
                dir_part = os.path.relpath(root, self.creds.folder)
                dir_part = "" if dir_part == "." else dir_part + "/"
                file_path = os.path.join(root, name)
                blob_path = dir_part + name
                self.upload_file(file_path, blob_path)

    def upload_file(self, source, dest):
        """Upload a single file to a path inside the container"""
        logging.info("Uploading %s to %s", source, dest)
        mime = magic.Magic(mime=True)
        with open(source, "rb") as data:
            if ".css" in source[-4:]:
                mimetype = ContentSettings(content_type="text/css")
            else:
                mimetype = ContentSettings(content_type=mime.from_file(source))
            self.containerClient.upload_blob(
                name=dest, data=data, content_settings=mimetype
            )



"""setup absu package"""
import setuptools


setuptools.setup(
    name="absu",
    version="0.0.1",
    author="Jonas C. Rabe aka Quisl",
    description=(
        "This module helps updating a static website hosted by Azure Blob "
        "Storage."
    ),
    packages=["absu", "absu.modules", "absu.modules.templates"],
    install_requires=[
        "azure-storage-blob==12.13.1",  # create container and upload files
        "python-magic-bin==0.4.14",  # set Mime Types automatically
        "azure-mgmt-resource>=18.0.0",  # create resources on azure
        "azure-mgmt-subscription>=3.1.1",  # get subscription
        "azure-identity>=1.5.0",  # connect to azure
        "azure-core==1.25.1",  # for azure exceptions
    ],
)



"""Provides a class that handles general communication with Azure"""
import logging
import os
import json
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.subscription import SubscriptionClient
from azure.identity import AzureCliCredential
from azure.core.exceptions import ClientAuthenticationError
from azure.mgmt.resource.resources.models import DeploymentMode, Deployment
from azure.core.exceptions import HttpResponseError
from .templates.storage import TEMPLATE


class AzureConnector:
    """This class handles Azure connections and can create/update resources"""

    def __init__(self):
        self.credential = AzureCliCredential()
        self.subscriptionId = self.get_subscription()
        self.location = "westeurope"
        self.resourceClient = ResourceManagementClient(
            self.credential, self.subscriptionId
        )

    def get_subscription(self):
        """returns a subscription id. executes az login if not logged in"""
        try:
            sub_client = SubscriptionClient(self.credential)
        except ClientAuthenticationError:
            logging.info("Not logged in, running az login")
            print("ROFL")
            process = os.popen("az login").read()
            sub_client = SubscriptionClient(self.credential)
        subs = [
            [sub.display_name, sub.subscription_id]
            for sub in sub_client.subscriptions.list()
        ]
        if len(subs) == 1:
            result = subs[0][1]
        else:
            print("Found multiple subscriptions...")
            for idx, sub in enumerate(subs):
                print(f"{str(idx+1)}: {str(sub)}")
            subscription_option = int(input("Select number: ")) - 1
            print(type(subscription_option))
            result = subs[subscription_option][1]
        return result

    def get_connection_string(self, storage, resourcegroup):
        """gets a connection string from"""
        return json.loads(
            os.popen(
                f"az storage account show-connection-string --name {storage} "
                f"--resource-group {resourcegroup} "
                f"--subscription {self.subscriptionId}"
            ).read()
        )["connectionString"]

    def get_storage_account_web_endpoint(self, storage, resourcegroup):
        """returns the address of the web endpoint"""
        return self.resourceClient.resources.get(
            resource_group_name=resourcegroup,
            resource_provider_namespace="Microsoft.Storage",
            api_version="2022-05-01",
            resource_name=storage,
            parent_resource_path="",
            resource_type="storageAccounts",
        ).__dict__["properties"]["primaryEndpoints"]["web"]

    def create_resource_group(self, resourcegroup):
        """creates a resource group"""
        rg_result = self.resourceClient.resource_groups.create_or_update(
            resourcegroup, {"location": self.location}
        )

        logging.debug(rg_result)

    def create_storage_account(self, resourcegroup, storage):
        """creates a storage (as defined in templates/storage.py)"""
        parameters = {
            "storageAccounts_blogdatastore_name": storage,
        }
        parameters = {k: {"value": v} for k, v in parameters.items()}
        deployment_properties = {
            "mode": DeploymentMode.incremental,
            "template": TEMPLATE,
            "parameters": parameters,
        }
        try:
            deployment_async_operation = self.resourceClient.deployments.begin_create_or_update(
                resource_group_name=resourcegroup,
                deployment_name="azure-sample",
                parameters=Deployment(
                    tags=None,
                    # location="westeurope",
                    properties=deployment_properties,
                ),
            )
            deployment_async_operation.wait()
            result = True
        except HttpResponseError:
            result = False
        return result
